@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2015.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2015. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -16,6 +16,7 @@
 #include <ucs/debug/log.h>
 #include <ucs/async/async.h>
 #include <ucs/sys/string.h>
+#include <uct/api/v2/uct_v2.h>
 
 
 #define PRINT_CAP(_name, _cap_flags, _max) \
@@ -55,6 +56,11 @@
         printf(" nsec\n"); \
     }
 
+#define PRINT_MD_MEM_TYPE(_strb, _md_attr, _mem_type, _cap) \
+    if ((_md_attr)._cap##_mem_types & UCS_BIT(_mem_type)) { \
+        ucs_string_buffer_appendf(_strb, UCS_PP_MAKE_STRING(_cap) ","); \
+    }
+
 
 static char *strduplower(const char *str)
 {
@@ -387,12 +393,14 @@ static void print_md_info(uct_component_h component,
                           ucs_config_print_flags_t print_flags,
                           const char *req_tl_name)
 {
+    UCS_STRING_BUFFER_ONSTACK(strb, 256);
     uct_tl_resource_desc_t *resources, tmp;
     unsigned resource_index, j, num_resources, count;
+    ucs_memory_type_t mem_type;
     ucs_status_t status;
     const char *tl_name;
     uct_md_config_t *md_config;
-    uct_md_attr_t md_attr;
+    uct_md_attr_v2_t md_attr;
     uct_md_h md;
 
     status = uct_md_config_read(component, NULL, NULL, &md_config);
@@ -431,7 +439,8 @@ static void print_md_info(uct_component_h component,
         }
     }
 
-    status = uct_md_query(md, &md_attr);
+    md_attr.field_mask = UINT64_MAX;
+    status             = uct_md_query_v2(md, &md_attr);
     if (status != UCS_OK) {
         printf("# < failed to query memory domain >\n");
         goto out_free_list;
@@ -439,27 +448,64 @@ static void print_md_info(uct_component_h component,
         printf("#\n");
         printf("# Memory domain: %s\n", md_name);
         printf("#     Component: %s\n", component_attr->name);
-        if (md_attr.cap.flags & UCT_MD_FLAG_ALLOC) {
+        if (md_attr.flags & UCT_MD_FLAG_ALLOC) {
             printf("#             allocate: %s\n",
-                   size_limit_to_str(0, md_attr.cap.max_alloc));
+                   size_limit_to_str(0, md_attr.max_alloc));
         }
-        if (md_attr.cap.flags & UCT_MD_FLAG_REG) {
-            printf("#             register: %s, cost: ",
-                   size_limit_to_str(0, md_attr.cap.max_reg));
+        if (md_attr.flags & UCT_MD_FLAG_REG) {
+            printf("#             register: %s",
+                   size_limit_to_str(0, md_attr.max_reg));
+            if (md_attr.flags & UCT_MD_FLAG_REG_DMABUF) {
+                printf(", dmabuf");
+            }
+            printf(", cost: ");
             PRINT_LINEAR_FUNC_NS(&md_attr.reg_cost);
         }
-        if (md_attr.cap.flags & UCT_MD_FLAG_NEED_RKEY) {
+        if (md_attr.flags & UCT_MD_FLAG_NEED_RKEY) {
             printf("#           remote key: %zu bytes\n", md_attr.rkey_packed_size);
         }
-        if (md_attr.cap.flags & UCT_MD_FLAG_NEED_MEMH) {
+        if (md_attr.flags & UCT_MD_FLAG_NEED_MEMH) {
             printf("#           local memory handle is required for zcopy\n");
         }
-        if (md_attr.cap.flags & UCT_MD_FLAG_RKEY_PTR) {
+        if (component_attr->flags & UCT_COMPONENT_FLAG_RKEY_PTR) {
             printf("#           rkey_ptr is supported\n");
         }
-        if (md_attr.cap.flags & UCT_MD_FLAG_INVALIDATE) {
+        if (md_attr.flags & UCT_MD_FLAG_INVALIDATE) {
             printf("#           memory invalidation is supported\n");
         }
+
+        ucs_memory_type_for_each(mem_type) {
+            if (!(UCS_BIT(mem_type) &
+                  (md_attr.reg_mem_types | md_attr.alloc_mem_types))) {
+                continue;
+            }
+
+            ucs_string_buffer_appendf(&strb, "%s",
+                                      ucs_memory_type_names[mem_type]);
+            if (!(UCS_BIT(mem_type) &
+                  (md_attr.access_mem_types | md_attr.alloc_mem_types |
+                   md_attr.reg_mem_types | md_attr.cache_mem_types |
+                   md_attr.detect_mem_types | md_attr.dmabuf_mem_types))) {
+                ucs_string_buffer_appendf(&strb, ", ");
+                continue;
+            }
+
+            ucs_string_buffer_appendf(&strb, " (");
+
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, access);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, alloc);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, reg_nonblock);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, reg);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, cache);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, detect);
+            PRINT_MD_MEM_TYPE(&strb, md_attr, mem_type, dmabuf);
+            ucs_string_buffer_rtrim(&strb, ",");
+
+            ucs_string_buffer_appendf(&strb, "), ");
+        }
+
+        ucs_string_buffer_rtrim(&strb, ", ");
+        printf("#         memory types: %s\n", ucs_string_buffer_cstr(&strb));
     }
 
     if (num_resources == 0) {
