@@ -89,7 +89,7 @@ int ADIOI_Calc_aggregator(ADIO_File fd,
         /* wkliao: implementation for file domain alignment
          * fd_start[] and fd_end[] have been aligned with file lock
          * boundaries when returned from ADIOI_Calc_file_domains() so cannot
-         * just use simple arithmatic as above */
+         * just use simple arithmetic as above */
         rank_index = 0;
         while (off > fd_end[rank_index])
             rank_index++;
@@ -166,7 +166,7 @@ void ADIOI_Calc_file_domains(ADIO_Offset * st_offsets, ADIO_Offset
     /* ceiling division as in HPF block distribution */
 
     /* Tweak the file domains so that no fd is smaller than a threshold.  We
-     * have to strike a balance between efficency and parallelism: somewhere
+     * have to strike a balance between efficiency and parallelism: somewhere
      * between 10k processes sending 32-byte requests and one process sending a
      * 320k request is a (system-dependent) sweet spot */
 
@@ -247,19 +247,19 @@ void ADIOI_Calc_file_domains(ADIO_Offset * st_offsets, ADIO_Offset
  * (including this one)
  */
 void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * len_list,
-                       int contig_access_count, ADIO_Offset
-                       min_st_offset, ADIO_Offset * fd_start,
+                       MPI_Count contig_access_count,
+                       ADIO_Offset min_st_offset, ADIO_Offset * fd_start,
                        ADIO_Offset * fd_end, ADIO_Offset fd_size,
                        int nprocs,
-                       int *count_my_req_procs_ptr,
-                       int **count_my_req_per_proc_ptr,
+                       MPI_Count *count_my_req_procs_ptr,
+                       MPI_Count **count_my_req_per_proc_ptr,
                        ADIOI_Access ** my_req_ptr, MPI_Aint ** buf_idx_ptr)
 /* Possibly reconsider if buf_idx's are ok as int's, or should they be aints/offsets?
    They are used as memory buffer indices so it seems like the 2G limit is in effect */
 {
-    int *count_my_req_per_proc, count_my_req_procs;
+    MPI_Count *count_my_req_per_proc, count_my_req_procs, l;
     MPI_Aint *buf_idx;
-    int i, l, proc;
+    int proc;
     size_t memLen;
     ADIO_Offset fd_len, rem_len, curr_idx, off, *ptr;
     ADIOI_Access *my_req;
@@ -268,7 +268,7 @@ void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * le
     MPE_Log_event(5024, 0, NULL);
 #endif
 
-    *count_my_req_per_proc_ptr = (int *) ADIOI_Calloc(nprocs, sizeof(int));
+    *count_my_req_per_proc_ptr = ADIOI_Calloc(nprocs, sizeof(MPI_Count));
     count_my_req_per_proc = *count_my_req_per_proc_ptr;
 /* count_my_req_per_proc[i] gives the no. of contig. requests of this
    process in process i's file domain. calloc initializes to zero.
@@ -282,13 +282,13 @@ void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * le
    without extra buffer. This can't be done if buftype is not contig. */
 
     /* initialize buf_idx to -1 */
-    for (i = 0; i < nprocs; i++)
+    for (int i = 0; i < nprocs; i++)
         buf_idx[i] = -1;
 
     /* one pass just to calculate how much space to allocate for my_req;
      * contig_access_count was calculated way back in ADIOI_Calc_my_off_len()
      */
-    for (i = 0; i < contig_access_count; i++) {
+    for (MPI_Count i = 0; i < contig_access_count; i++) {
         /* short circuit offset/len processing if len == 0
          *      (zero-byte  read/write */
         if (len_list[i] == 0)
@@ -330,13 +330,13 @@ void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * le
      * make 'lens' point to the over-allocated part
      */
     memLen = 0;
-    for (i = 0; i < nprocs; i++)
+    for (int i = 0; i < nprocs; i++)
         memLen += count_my_req_per_proc[i];
     ptr = (ADIO_Offset *) ADIOI_Malloc(memLen * 2 * sizeof(ADIO_Offset));
     my_req[0].offsets = ptr;
 
     count_my_req_procs = 0;
-    for (i = 0; i < nprocs; i++) {
+    for (int i = 0; i < nprocs; i++) {
         if (count_my_req_per_proc[i]) {
             my_req[i].offsets = ptr;
             ptr += count_my_req_per_proc[i];
@@ -350,7 +350,7 @@ void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * le
 
 /* now fill in my_req */
     curr_idx = 0;
-    for (i = 0; i < contig_access_count; i++) {
+    for (MPI_Count i = 0; i < contig_access_count; i++) {
         /* short circuit offset/len processing if len == 0
          *      (zero-byte  read/write */
         if (len_list[i] == 0)
@@ -420,7 +420,7 @@ void ADIOI_Calc_my_req(ADIO_File fd, ADIO_Offset * offset_list, ADIO_Offset * le
 #endif
 }
 
-void ADIOI_Free_my_req(int nprocs, int *count_my_req_per_proc,
+void ADIOI_Free_my_req(int nprocs, MPI_Count *count_my_req_per_proc,
                        ADIOI_Access * my_req, MPI_Aint * buf_idx)
 {
     ADIOI_Free(count_my_req_per_proc);
@@ -429,12 +429,12 @@ void ADIOI_Free_my_req(int nprocs, int *count_my_req_per_proc,
     ADIOI_Free(buf_idx);
 }
 
-void ADIOI_Calc_others_req(ADIO_File fd, int count_my_req_procs,
-                           int *count_my_req_per_proc,
+void ADIOI_Calc_others_req(ADIO_File fd, MPI_Count count_my_req_procs,
+                           MPI_Count *count_my_req_per_proc,
                            ADIOI_Access * my_req,
                            int nprocs, int myrank,
-                           int *count_others_req_procs_ptr,
-                           int **count_others_req_per_proc_ptr, ADIOI_Access ** others_req_ptr)
+                           MPI_Count *count_others_req_procs_ptr,
+                           MPI_Count **count_others_req_per_proc_ptr, ADIOI_Access ** others_req_ptr)
 {
 /* determine what requests of other processes lie in this process's
    file domain */
@@ -444,22 +444,22 @@ void ADIOI_Calc_others_req(ADIO_File fd, int count_my_req_procs,
    count_others_req_per_proc[i] indicates how many separate contiguous
    requests of proc. i lie in this process's file domain. */
 
-    int *count_others_req_per_proc, count_others_req_procs;
+    MPI_Count *count_others_req_per_proc, count_others_req_procs;
     int i, j;
     MPI_Request *requests;
     ADIOI_Access *others_req;
     size_t memLen;
     ADIO_Offset *ptr;
-    MPI_Aint *mem_ptrs;
+    MPI_Count *mem_ptrs;
 
 /* first find out how much to send/recv and from/to whom */
 #ifdef AGGREGATION_PROFILE
     MPE_Log_event(5026, 0, NULL);
 #endif
-    count_others_req_per_proc = (int *) ADIOI_Malloc(nprocs * sizeof(int));
+    count_others_req_per_proc = ADIOI_Malloc(nprocs * sizeof(MPI_Count));
 
-    MPI_Alltoall(count_my_req_per_proc, 1, MPI_INT,
-                 count_others_req_per_proc, 1, MPI_INT, fd->comm);
+    MPI_Alltoall(count_my_req_per_proc, 1, MPI_COUNT,
+                 count_others_req_per_proc, 1, MPI_COUNT, fd->comm);
 
     *others_req_ptr = (ADIOI_Access *) ADIOI_Malloc(nprocs * sizeof(ADIOI_Access));
     others_req = *others_req_ptr;
@@ -468,7 +468,7 @@ void ADIOI_Calc_others_req(ADIO_File fd, int count_my_req_procs,
     for (i = 0; i < nprocs; i++)
         memLen += count_others_req_per_proc[i];
     ptr = (ADIO_Offset *) ADIOI_Malloc(memLen * 2 * sizeof(ADIO_Offset));
-    mem_ptrs = (MPI_Aint *) ADIOI_Malloc(memLen * sizeof(MPI_Aint));
+    mem_ptrs = ADIOI_Malloc(memLen * sizeof(*mem_ptrs));
     others_req[0].offsets = ptr;
     others_req[0].mem_ptrs = mem_ptrs;
 
@@ -497,14 +497,14 @@ void ADIOI_Calc_others_req(ADIO_File fd, int count_my_req_procs,
     j = 0;
     for (i = 0; i < nprocs; i++) {
         if (others_req[i].count) {
-            MPI_Irecv(others_req[i].offsets, 2 * others_req[i].count,
+            MPI_Irecv_c(others_req[i].offsets, 2 * others_req[i].count,
                       ADIO_OFFSET, i, i + myrank, fd->comm, &requests[j++]);
         }
     }
 
     for (i = 0; i < nprocs; i++) {
         if (my_req[i].count) {
-            MPI_Isend(my_req[i].offsets, 2 * my_req[i].count,
+            MPI_Isend_c(my_req[i].offsets, 2 * my_req[i].count,
                       ADIO_OFFSET, i, i + myrank, fd->comm, &requests[j++]);
         }
     }
@@ -527,7 +527,7 @@ void ADIOI_Calc_others_req(ADIO_File fd, int count_my_req_procs,
 #endif
 }
 
-void ADIOI_Free_others_req(int nprocs, int *count_others_req_per_proc, ADIOI_Access * others_req)
+void ADIOI_Free_others_req(int nprocs, MPI_Count *count_others_req_per_proc, ADIOI_Access * others_req)
 {
     ADIOI_Free(count_others_req_per_proc);
     ADIOI_Free(others_req[0].offsets);
@@ -549,10 +549,10 @@ void ADIOI_Icalc_others_req(ADIOI_NBC_Request * nbc_req, int *error_code)
 #ifdef AGGREGATION_PROFILE
     MPE_Log_event(5026, 0, NULL);
 #endif
-    vars->count_others_req_per_proc = (int *) ADIOI_Malloc(vars->nprocs * sizeof(int));
+    vars->count_others_req_per_proc = ADIOI_Malloc(vars->nprocs * sizeof(MPI_Count));
 
-    *error_code = MPI_Ialltoall(vars->count_my_req_per_proc, 1, MPI_INT,
-                                vars->count_others_req_per_proc, 1, MPI_INT, vars->fd->comm,
+    *error_code = MPI_Ialltoall(vars->count_my_req_per_proc, 1, MPI_COUNT,
+                                vars->count_others_req_per_proc, 1, MPI_COUNT, vars->fd->comm,
                                 &vars->req1);
 
     if (nbc_req->rdwr == ADIOI_READ) {
@@ -567,7 +567,7 @@ void ADIOI_Icalc_others_req_main(ADIOI_NBC_Request * nbc_req, int *error_code)
 {
     ADIOI_Icalc_others_req_vars *vars = nbc_req->cor_vars;
     ADIO_File fd = vars->fd;
-    int count_my_req_procs = vars->count_my_req_procs;
+    MPI_Count count_my_req_procs = vars->count_my_req_procs;
     ADIOI_Access *my_req = vars->my_req;
     int nprocs = vars->nprocs;
     int myrank = vars->myrank;
@@ -581,13 +581,13 @@ void ADIOI_Icalc_others_req_main(ADIOI_NBC_Request * nbc_req, int *error_code)
      * count_others_req_per_proc[i] indicates how many separate contiguous
      * requests of proc. i lie in this process's file domain. */
 
-    int *count_others_req_per_proc = vars->count_others_req_per_proc;
-    int count_others_req_procs;
+    MPI_Count *count_others_req_per_proc = vars->count_others_req_per_proc;
+    MPI_Count count_others_req_procs;
     int i, j;
     ADIOI_Access *others_req;
     size_t memLen;
     ADIO_Offset *ptr;
-    MPI_Aint *mem_ptrs;
+    MPI_Count *mem_ptrs;
 
     *others_req_ptr = (ADIOI_Access *) ADIOI_Malloc(nprocs * sizeof(ADIOI_Access));
     others_req = *others_req_ptr;
@@ -596,7 +596,7 @@ void ADIOI_Icalc_others_req_main(ADIOI_NBC_Request * nbc_req, int *error_code)
     for (i = 0; i < nprocs; i++)
         memLen += count_others_req_per_proc[i];
     ptr = (ADIO_Offset *) ADIOI_Malloc(memLen * 2 * sizeof(ADIO_Offset));
-    mem_ptrs = (MPI_Aint *) ADIOI_Malloc(memLen * sizeof(MPI_Aint));
+    mem_ptrs = ADIOI_Malloc(memLen * sizeof(*mem_ptrs));
     others_req[0].offsets = ptr;
     others_req[0].mem_ptrs = mem_ptrs;
 
@@ -626,14 +626,14 @@ void ADIOI_Icalc_others_req_main(ADIOI_NBC_Request * nbc_req, int *error_code)
     j = 0;
     for (i = 0; i < nprocs; i++) {
         if (others_req[i].count) {
-            MPI_Irecv(others_req[i].offsets, 2 * others_req[i].count,
+            MPI_Irecv_c(others_req[i].offsets, 2 * others_req[i].count,
                       ADIO_OFFSET, i, i + myrank, fd->comm, &vars->req2[j++]);
         }
     }
 
     for (i = 0; i < nprocs; i++) {
         if (my_req[i].count) {
-            MPI_Isend(my_req[i].offsets, 2 * my_req[i].count,
+            MPI_Isend_c(my_req[i].offsets, 2 * my_req[i].count,
                       ADIO_OFFSET, i, i + myrank, fd->comm, &vars->req2[j++]);
         }
     }
