@@ -1,8 +1,8 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2001 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include "mpi.h"
 #include <stdio.h>
 #include <string.h>
@@ -78,7 +78,11 @@ int main(int argc, char **argv)
     if (errcode != MPI_SUCCESS) {
         handle_error(errcode, "MPI_File_open");
     }
-    MPI_File_set_view(fh, 0, MPI_INT, MPI_INT, "native", MPI_INFO_NULL);
+    errcode = MPI_File_set_view(fh, 0, MPI_INT, MPI_INT, "native", MPI_INFO_NULL);
+    if (errcode != MPI_SUCCESS) {
+        handle_error(errcode, "MPI_File_set_view");
+    }
+
     errcode = MPI_File_iwrite(fh, buf, nints, MPI_INT, &request);
     if (errcode != MPI_SUCCESS) {
         handle_error(errcode, "MPI_File_iwrite");
@@ -100,7 +104,10 @@ int main(int argc, char **argv)
         handle_error(errcode, "MPI_File_open");
     }
 
-    MPI_File_set_view(fh, 0, MPI_INT, MPI_INT, "native", MPI_INFO_NULL);
+    errcode = MPI_File_set_view(fh, 0, MPI_INT, MPI_INT, "native", MPI_INFO_NULL);
+    if (errcode != MPI_SUCCESS) {
+        handle_error(errcode, "MPI_File_set_view");
+    }
     errcode = MPI_File_iread(fh, buf, nints, MPI_INT, &request);
     if (errcode != MPI_SUCCESS) {
         handle_error(errcode, "MPI_File_open");
@@ -136,5 +143,5 @@ int main(int argc, char **argv)
     free(tmp);
 
     MPI_Finalize();
-    return 0;
+    return (toterrs > 0);
 }
