@@ -24,7 +24,7 @@
 MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_recv_iov(void *buf, MPI_Aint count, size_t data_sz,      /* data_sz passed in here for reusing */
                                                 int rank, uint64_t match_bits, uint64_t mask_bits,
                                                 MPIR_Comm * comm, MPIR_Context_id_t context_id,
-                                                MPIDI_av_entry_t * addr, int vni_src, int vni_dst,
+                                                MPIDI_av_entry_t * addr, int vci_src, int vci_dst,
                                                 MPIR_Request * rreq,
                                                 MPIR_Datatype * dt_ptr, uint64_t flags)
 {
@@ -32,29 +32,25 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_recv_iov(void *buf, MPI_Aint count, size_
     struct iovec *originv = NULL;
     struct fi_msg_tagged msg;
     MPI_Aint num_contig, size;
-    int vni_remote = vni_src;
-    int vni_local = vni_dst;
-    int ctx_idx = 0;
-    int sender_nic = 0, receiver_nic = 0;
+    int vci_remote = vci_src;
+    int vci_local = vci_dst;
+    int ctx_idx;
+    int receiver_nic;
 
     MPIR_FUNC_ENTER;
 
     /* if we cannot fit the entire data into a single IOV array,
      * fallback to pack */
-    MPIR_Typerep_iov_len(count, MPIDI_OFI_REQUEST(rreq, datatype), dt_ptr->size * count,
-                         &num_contig);
+    MPIR_Typerep_get_iov_len(count, MPIDI_OFI_REQUEST(rreq, datatype), &num_contig);
     if (num_contig > MPIDI_OFI_global.rx_iov_limit)
         goto unpack;
 
     /* Calculate the correct NICs. */
-    sender_nic =
-        MPIDI_OFI_multx_sender_nic_index(comm, comm->recvcontext_id, rank, comm->rank,
-                                         MPIDI_OFI_init_get_tag(match_bits));
     receiver_nic =
         MPIDI_OFI_multx_receiver_nic_index(comm, comm->recvcontext_id, rank, comm->rank,
                                            MPIDI_OFI_init_get_tag(match_bits));
     MPIDI_OFI_REQUEST(rreq, nic_num) = receiver_nic;
-    ctx_idx = MPIDI_OFI_get_ctx_index(comm, vni_dst, MPIDI_OFI_REQUEST(rreq, nic_num));
+    ctx_idx = MPIDI_OFI_get_ctx_index(vci_dst, MPIDI_OFI_REQUEST(rreq, nic_num));
 
     if (!flags) {
         flags = FI_COMPLETION;
@@ -65,10 +61,10 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_recv_iov(void *buf, MPI_Aint count, size_
     MPIDI_OFI_REQUEST(rreq, noncontig.nopack) = MPL_malloc(size, MPL_MEM_BUFFER);
     memset(MPIDI_OFI_REQUEST(rreq, noncontig.nopack), 0, size);
 
-    MPI_Aint actual_iov_len, actual_iov_bytes;
-    MPIR_Typerep_to_iov(buf, count, MPIDI_OFI_REQUEST(rreq, datatype), 0,
-                        MPIDI_OFI_REQUEST(rreq, noncontig.nopack), num_contig, dt_ptr->size * count,
-                        &actual_iov_len, &actual_iov_bytes);
+    MPI_Aint actual_iov_len;
+    MPIR_Typerep_to_iov_offset(buf, count, MPIDI_OFI_REQUEST(rreq, datatype), 0,
+                               MPIDI_OFI_REQUEST(rreq, noncontig.nopack), num_contig,
+                               &actual_iov_len);
     assert(num_contig == actual_iov_len);
 
     originv = &(MPIDI_OFI_REQUEST(rreq, noncontig.nopack[0]));
@@ -89,12 +85,17 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_recv_iov(void *buf, MPI_Aint count, size_
     msg.ignore = mask_bits;
     msg.context = (void *) &(MPIDI_OFI_REQUEST(rreq, context));
     msg.data = 0;
-    msg.addr =
-        (MPI_ANY_SOURCE == rank) ? FI_ADDR_UNSPEC : MPIDI_OFI_av_to_phys(addr, sender_nic,
-                                                                         vni_local, vni_remote);
+    if (MPI_ANY_SOURCE == rank) {
+        msg.addr = FI_ADDR_UNSPEC;
+    } else {
+        int sender_nic = MPIDI_OFI_multx_sender_nic_index(comm, comm->recvcontext_id,
+                                                          rank, comm->rank,
+                                                          MPIDI_OFI_init_get_tag(match_bits));
+        msg.addr = MPIDI_OFI_av_to_phys(addr, sender_nic, vci_remote);
+    }
 
-    MPIDI_OFI_CALL_RETRY(fi_trecvmsg(MPIDI_OFI_global.ctx[ctx_idx].rx, &msg, flags), vni_local,
-                         trecv, FALSE);
+    MPIDI_OFI_CALL_RETRY(fi_trecvmsg(MPIDI_OFI_global.ctx[ctx_idx].rx, &msg, flags), vci_local,
+                         trecv);
 
   fn_exit:
     MPIR_FUNC_EXIT;
@@ -115,7 +116,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
                                                 int tag,
                                                 MPIR_Comm * comm,
                                                 int context_offset,
-                                                MPIDI_av_entry_t * addr, int vni_src, int vni_dst,
+                                                MPIDI_av_entry_t * addr, int vci_src, int vci_dst,
                                                 MPIR_Request ** request, int mode, uint64_t flags)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -129,23 +130,18 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
     struct fi_msg_tagged msg;
     char *recv_buf;
     bool force_gpu_pack = false;
-    int vni_remote = vni_src;
-    int vni_local = vni_dst;
-    int sender_nic = 0, receiver_nic = 0;
-    int ctx_idx = 0;
+    int vci_remote = vci_src;
+    int vci_local = vci_dst;
+    int receiver_nic;
+    int ctx_idx;
+    bool register_mem = false;
+    struct fid_mr *mr = NULL;
+    void *desc = NULL;
 
     MPIR_FUNC_ENTER;
 
     if (mode == MPIDI_OFI_ON_HEAP) {    /* Branch should compile out */
-#ifdef MPIDI_CH4_USE_WORK_QUEUES
-        /* TODO: what cases when *request is NULL under workq? */
-        if (*request) {
-            MPIR_Request_add_ref(*request);
-        } else
-#endif
-        {
-            MPIDI_OFI_REQUEST_CREATE(*request, MPIR_REQUEST_KIND__RECV, vni_dst);
-        }
+        MPIDI_OFI_REQUEST_CREATE(*request, MPIR_REQUEST_KIND__RECV, vci_dst);
         rreq = *request;
 
         /* Need to set the source to UNDEFINED for anysource matching */
@@ -164,40 +160,53 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
     }
 
     /* Calculate the correct NICs. */
-    sender_nic =
-        MPIDI_OFI_multx_sender_nic_index(comm, comm->recvcontext_id, rank, comm->rank, tag);
     receiver_nic =
         MPIDI_OFI_multx_receiver_nic_index(comm, comm->recvcontext_id, rank, comm->rank, tag);
     MPIDI_OFI_REQUEST(rreq, nic_num) = receiver_nic;
-    ctx_idx = MPIDI_OFI_get_ctx_index(comm, vni_dst, MPIDI_OFI_REQUEST(rreq, nic_num));
+    ctx_idx = MPIDI_OFI_get_ctx_index(vci_dst, MPIDI_OFI_REQUEST(rreq, nic_num));
 
-    match_bits = MPIDI_OFI_init_recvtag(&mask_bits, context_id, tag);
+    match_bits = MPIDI_OFI_init_recvtag(&mask_bits, context_id, rank, tag);
 
     MPIDI_Datatype_get_info(count, datatype, dt_contig, data_sz, dt_ptr, dt_true_lb);
     MPIDI_OFI_REQUEST(rreq, datatype) = datatype;
     MPIR_Datatype_add_ref_if_not_builtin(datatype);
 
     recv_buf = MPIR_get_contig_ptr(buf, dt_true_lb);
-    MPL_pointer_attr_t attr;
+    MPL_pointer_attr_t attr = {.type = MPL_GPU_POINTER_UNREGISTERED_HOST };
     MPIR_GPU_query_pointer_attr(recv_buf, &attr);
-    if (data_sz && attr.type == MPL_GPU_POINTER_DEV) {
-        if (!MPIDI_OFI_ENABLE_HMEM) {
+
+    if (MPIDI_OFI_ENABLE_HMEM && data_sz >= MPIR_CVAR_CH4_OFI_GPU_RDMA_THRESHOLD &&
+        MPIDI_OFI_ENABLE_MR_HMEM && dt_contig) {
+        if (attr.type == MPL_GPU_POINTER_DEV) {
+            register_mem = true;
+        }
+    }
+
+    if (data_sz && MPL_gpu_query_pointer_is_dev(recv_buf, &attr)) {
+        MPIDI_OFI_register_am_bufs();
+        if (!MPIDI_OFI_ENABLE_HMEM || !dt_contig || (MPIDI_OFI_ENABLE_MR_HMEM && !register_mem)) {
             /* FIXME: at this point, GPU data takes host-buffer staging
              * path for the whole chunk. For large memory size, pipeline
              * transfer should be applied. */
             force_gpu_pack = true;
-            dt_contig = 0;
         }
     }
 
-    if (!dt_contig) {
+    if (register_mem) {
+        MPIDI_OFI_register_memory_and_bind(recv_buf, data_sz, &attr, ctx_idx, &mr);
+        if (mr != NULL) {
+            desc = fi_mr_desc(mr);
+        }
+    }
+
+    if (!dt_contig || force_gpu_pack) {
         if (MPIDI_OFI_ENABLE_PT2PT_NOPACK && !force_gpu_pack &&
             ((data_sz < MPIDI_OFI_global.max_msg_size && !MPIDI_OFI_COMM(comm).enable_striping) ||
              (data_sz < MPIDI_OFI_global.stripe_threshold &&
               MPIDI_OFI_COMM(comm).enable_striping))) {
             mpi_errno =
                 MPIDI_OFI_recv_iov(buf, count, data_sz, rank, match_bits, mask_bits, comm,
-                                   context_id, addr, vni_src, vni_dst, rreq, dt_ptr, flags);
+                                   context_id, addr, vci_src, vci_dst, rreq, dt_ptr, flags);
             if (mpi_errno == MPI_SUCCESS)       /* Receive posted using iov */
                 goto fn_exit;
             else if (mpi_errno != MPIDI_OFI_RECV_NEEDS_UNPACK)
@@ -208,6 +217,72 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
              * path below. Simply falling through. */
             mpi_errno = MPI_SUCCESS;    /* Reset error code */
         }
+
+        if (force_gpu_pack && MPIR_CVAR_CH4_OFI_ENABLE_GPU_PIPELINE &&
+            data_sz >= MPIR_CVAR_CH4_OFI_GPU_PIPELINE_THRESHOLD) {
+            /* Pipeline path */
+            MPL_atomic_relaxed_store_int(&MPIDI_OFI_REQUEST(rreq, util_id), context_id);
+            MPIDI_OFI_REQUEST(rreq, event_id) = MPIDI_OFI_EVENT_RECV_GPU_PIPELINE_INIT;
+            /* Only post first recv with pipeline chunk size. */
+            char *host_buf = NULL;
+            MPIDU_genq_private_pool_force_alloc_cell(MPIDI_OFI_global.gpu_pipeline_recv_pool,
+                                                     (void **) &host_buf);
+            MPIR_ERR_CHKANDJUMP1(host_buf == NULL, mpi_errno,
+                                 MPI_ERR_OTHER, "**nomem", "**nomem %s",
+                                 "Pipeline Init recv alloc");
+
+            fi_addr_t remote_addr;
+            if (MPI_ANY_SOURCE == rank)
+                remote_addr = FI_ADDR_UNSPEC;
+            else {
+                int sender_nic =
+                    MPIDI_OFI_multx_sender_nic_index(comm, comm->recvcontext_id, rank, comm->rank,
+                                                     MPIDI_OFI_init_get_tag(match_bits));
+                remote_addr = MPIDI_OFI_av_to_phys(addr, sender_nic, vci_remote);
+            }
+
+            /* Save pipeline information. */
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.offset) = 0;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.is_sync) = false;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.remote_addr) = remote_addr;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.vci_local) = vci_local;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.match_bits) = match_bits;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.mask_bits) = mask_bits;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.data_sz) = data_sz;
+            MPIDI_OFI_REQUEST(rreq, pipeline_info.ctx_idx) = ctx_idx;
+
+            /* Save original buf, datatype and count */
+            MPIDI_OFI_REQUEST(rreq, noncontig.pack.pack_buffer) = host_buf;
+            MPIDI_OFI_REQUEST(rreq, noncontig.pack.buf) = buf;
+            MPIDI_OFI_REQUEST(rreq, noncontig.pack.count) = count;
+            MPIDI_OFI_REQUEST(rreq, noncontig.pack.datatype) = datatype;
+
+            if (rreq->comm == NULL) {
+                rreq->comm = comm;
+                MPIR_Comm_add_ref(comm);
+            }
+
+            MPIDI_OFI_gpu_pipeline_request *chunk_req;
+            chunk_req = (MPIDI_OFI_gpu_pipeline_request *)
+                MPL_malloc(sizeof(MPIDI_OFI_gpu_pipeline_request), MPL_MEM_BUFFER);
+            MPIR_ERR_CHKANDJUMP1(chunk_req == NULL, mpi_errno,
+                                 MPI_ERR_OTHER, "**nomem", "**nomem %s", "Recv chunk_req alloc");
+            chunk_req->event_id = MPIDI_OFI_EVENT_RECV_GPU_PIPELINE_INIT;
+            chunk_req->parent = rreq;
+            chunk_req->buf = host_buf;
+            int ret = 0;
+            ret = fi_trecv(MPIDI_OFI_global.ctx[ctx_idx].rx,
+                           host_buf,
+                           MPIR_CVAR_CH4_OFI_GPU_PIPELINE_BUFFER_SZ,
+                           NULL, remote_addr, match_bits, mask_bits, (void *) &chunk_req->context);
+            if (MPIDI_OFI_global.gpu_recv_queue || !host_buf || ret != 0) {
+                MPIDI_OFI_gpu_pending_recv_t *recv_task =
+                    MPIDI_OFI_create_recv_task(chunk_req, 0, -1);
+                DL_APPEND(MPIDI_OFI_global.gpu_recv_queue, recv_task);
+            }
+            goto fn_exit;
+        }
+
         /* Unpack */
         MPIDI_OFI_REQUEST(rreq, event_id) = MPIDI_OFI_EVENT_RECV_PACK;
         MPIDI_OFI_REQUEST(rreq, noncontig.pack.pack_buffer) = MPL_malloc(data_sz, MPL_MEM_OTHER);
@@ -215,6 +290,16 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
                              MPI_ERR_OTHER, "**nomem", "**nomem %s", "Recv Pack Buffer alloc");
         recv_buf = MPIDI_OFI_REQUEST(rreq, noncontig.pack.pack_buffer);
         MPIDI_OFI_REQUEST(rreq, noncontig.pack.buf) = buf;
+#ifdef MPL_HAVE_ZE
+        if (dt_contig && attr.type == MPL_GPU_POINTER_DEV) {
+            int mpl_err = MPL_SUCCESS;
+            void *ptr;
+            mpl_err = MPL_ze_mmap_device_pointer(buf, &attr.device_attr, attr.device, &ptr);
+            MPIR_ERR_CHKANDJUMP(mpl_err != MPL_SUCCESS, mpi_errno, MPI_ERR_OTHER,
+                                "**mpl_ze_mmap_device_ptr");
+            MPIDI_OFI_REQUEST(rreq, noncontig.pack.buf) = ptr;
+        }
+#endif
         MPIDI_OFI_REQUEST(rreq, noncontig.pack.count) = count;
         MPIDI_OFI_REQUEST(rreq, noncontig.pack.datatype) = datatype;
     } else {
@@ -243,18 +328,22 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
         MPIDI_OFI_REQUEST(rreq, event_id) = MPIDI_OFI_EVENT_RECV;
 
     if (!flags) {
+        fi_addr_t sender_addr;
+        if (MPI_ANY_SOURCE == rank) {
+            sender_addr = FI_ADDR_UNSPEC;
+        } else {
+            int sender_nic = MPIDI_OFI_multx_sender_nic_index(comm, comm->recvcontext_id,
+                                                              rank, comm->rank,
+                                                              MPIDI_OFI_init_get_tag(match_bits));
+            sender_addr = MPIDI_OFI_av_to_phys(addr, sender_nic, vci_remote);
+        }
         MPIDI_OFI_CALL_RETRY(fi_trecv(MPIDI_OFI_global.ctx[ctx_idx].rx,
-                                      recv_buf,
-                                      data_sz,
-                                      NULL,
-                                      (MPI_ANY_SOURCE == rank) ? FI_ADDR_UNSPEC :
-                                      MPIDI_OFI_av_to_phys(addr, sender_nic, vni_local, vni_remote),
-                                      match_bits, mask_bits,
-                                      (void *) &(MPIDI_OFI_REQUEST(rreq, context))), vni_local,
-                             trecv, FALSE);
+                                      recv_buf, data_sz, desc, sender_addr, match_bits, mask_bits,
+                                      (void *) &(MPIDI_OFI_REQUEST(rreq, context))), vci_local,
+                             trecv);
     } else {
         msg.msg_iov = &MPIDI_OFI_REQUEST(rreq, util.iov);
-        msg.desc = NULL;
+        msg.desc = desc;
         msg.iov_count = 1;
         msg.tag = match_bits;
         msg.ignore = mask_bits;
@@ -262,8 +351,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
         msg.data = 0;
         msg.addr = FI_ADDR_UNSPEC;
 
-        MPIDI_OFI_CALL_RETRY(fi_trecvmsg(MPIDI_OFI_global.ctx[ctx_idx].rx, &msg, flags), vni_local,
-                             trecv, FALSE);
+        MPIDI_OFI_CALL_RETRY(fi_trecvmsg(MPIDI_OFI_global.ctx[ctx_idx].rx, &msg, flags), vci_local,
+                             trecv);
     }
 
   fn_exit:
@@ -274,20 +363,22 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_irecv(void *buf,
 }
 
 /* Common macro used by all MPIDI_NM_mpi_irecv routines to facilitate tuning */
-#define MPIDI_OFI_RECV_VNIS(vni_src_, vni_dst_) \
+#define MPIDI_OFI_RECV_VNIS(vci_src_, vci_dst_) \
     do { \
         if (*request != NULL) { \
             /* workq path  or collectives */ \
-            vni_src_ = 0; \
-            vni_dst_ = 0; \
+            vci_src_ = 0; \
+            vci_dst_ = 0; \
         } else { \
             /* NOTE: hashing is based on target rank */ \
-            vni_src_ = MPIDI_OFI_get_vni(SRC_VCI_FROM_RECVER, comm, rank, comm->rank, tag); \
-            vni_dst_ = MPIDI_OFI_get_vni(DST_VCI_FROM_RECVER, comm, rank, comm->rank, tag); \
+            MPIDI_EXPLICIT_VCIS(comm, attr, rank, comm->rank, vci_src_, vci_dst_); \
+            if (vci_src_ == 0 && vci_dst_ == 0) { \
+                vci_src_ = MPIDI_get_vci(SRC_VCI_FROM_RECVER, comm, rank, comm->rank, tag); \
+                vci_dst_ = MPIDI_get_vci(DST_VCI_FROM_RECVER, comm, rank, comm->rank, tag); \
+            } \
         } \
     } while (0)
 
-
 MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_imrecv(void *buf,
                                                  MPI_Aint count,
                                                  MPI_Datatype datatype, MPIR_Request * message)
@@ -297,19 +388,19 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_imrecv(void *buf,
     MPIDI_av_entry_t *av;
     MPIR_FUNC_ENTER;
 
+    int vci = MPIDI_Request_get_vci(message);
+    MPIDI_OFI_THREAD_CS_ENTER_VCI_OPTIONAL(vci);
     if (!MPIDI_OFI_ENABLE_TAGGED) {
         mpi_errno = MPIDIG_mpi_imrecv(buf, count, datatype, message);
     } else {
         rreq = message;
-        int vci = MPIDI_Request_get_vci(rreq);
-        MPIDI_OFI_THREAD_CS_ENTER_VCI_OPTIONAL(vci);
         av = MPIDIU_comm_rank_to_av(rreq->comm, message->status.MPI_SOURCE);
-        /* FIXME: need get vni_src in the request */
+        /* FIXME: need get vci_src in the request */
         mpi_errno = MPIDI_OFI_do_irecv(buf, count, datatype, message->status.MPI_SOURCE,
                                        message->status.MPI_TAG, rreq->comm, 0, av, 0, vci,
                                        &rreq, MPIDI_OFI_USE_EXISTING, FI_CLAIM | FI_COMPLETION);
-        MPIDI_OFI_THREAD_CS_EXIT_VCI_OPTIONAL(vci);
     }
+    MPIDI_OFI_THREAD_CS_EXIT_VCI_OPTIONAL(vci);
 
     MPIR_FUNC_EXIT;
     return mpi_errno;
@@ -320,27 +411,44 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_irecv(void *buf,
                                                 MPI_Datatype datatype,
                                                 int rank,
                                                 int tag,
-                                                MPIR_Comm * comm, int context_offset,
+                                                MPIR_Comm * comm, int attr,
                                                 MPIDI_av_entry_t * addr, MPIR_Request ** request,
                                                 MPIR_Request * partner)
 {
     int mpi_errno = MPI_SUCCESS;
     MPIR_FUNC_ENTER;
 
-    int vni_src, vni_dst;
-    MPIDI_OFI_RECV_VNIS(vni_src, vni_dst);
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
+    int vci_src, vci_dst;
+    MPIDI_OFI_RECV_VNIS(vci_src, vci_dst);
+
     /* For anysource recv, we may be called while holding the vci lock of shm request (to
-     * prevent shm progress). Therefore, recursive locking is allowed here */
+     * prevent shm progress). */
+    int need_cs;
+#ifdef MPIDI_CH4_DIRECT_NETMOD
+    need_cs = true;
+#else
+    need_cs = (rank != MPI_ANY_SOURCE);
+#endif
+
+    if (need_cs) {
+        MPIDI_OFI_THREAD_CS_ENTER_VCI_OPTIONAL(vci_dst);
+    } else {
+#ifdef MPICH_DEBUG_MUTEX
+        MPID_THREAD_ASSERT_IN_CS(VCI, MPIDI_VCI(vci_dst).lock);
+#endif
+    }
     if (!MPIDI_OFI_ENABLE_TAGGED) {
         mpi_errno = MPIDIG_mpi_irecv(buf, count, datatype, rank, tag, comm, context_offset,
-                                     vni_dst, request, 0, partner);
+                                     vci_dst, request, 0, partner);
     } else {
-        MPIDI_OFI_THREAD_CS_ENTER_REC_VCI_OPTIONAL(vni_dst);
         mpi_errno = MPIDI_OFI_do_irecv(buf, count, datatype, rank, tag, comm,
-                                       context_offset, addr, vni_src, vni_dst, request,
+                                       context_offset, addr, vci_src, vci_dst, request,
                                        MPIDI_OFI_ON_HEAP, 0ULL);
         MPIDI_REQUEST_SET_LOCAL(*request, 0, partner);
-        MPIDI_OFI_THREAD_CS_EXIT_VCI_OPTIONAL(vni_dst);
+    }
+    if (need_cs) {
+        MPIDI_OFI_THREAD_CS_EXIT_VCI_OPTIONAL(vci_dst);
     }
 
     MPIR_FUNC_EXIT;
@@ -351,39 +459,32 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_cancel_recv(MPIR_Request * rreq, bool
 {
 
     int mpi_errno = MPI_SUCCESS;
-    ssize_t ret;
     MPIR_FUNC_ENTER;
 
-    int vni = MPIDI_Request_get_vci(rreq);
-    int ctx_idx = MPIDI_OFI_get_ctx_index(rreq->comm, vni, MPIDI_OFI_REQUEST(rreq, nic_num));
+    int vci = MPIDI_Request_get_vci(rreq);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, MPIDI_OFI_REQUEST(rreq, nic_num));
 
     if (!MPIDI_OFI_ENABLE_TAGGED) {
         mpi_errno = MPIDIG_mpi_cancel_recv(rreq);
         goto fn_exit;
     }
 
-    /* Not using the OFI_CALL macro because there are error cases here that we want to catch */
-    ret = fi_cancel((fid_t) MPIDI_OFI_global.ctx[ctx_idx].rx, &(MPIDI_OFI_REQUEST(rreq, context)));
-    if (ret == -FI_ENOENT) {
-        /* The context was not found inside libfabric which means it was matched previously and has
-         * already been handled. Note that it is impossible to tell the difference in this case
-         * between a request that was previously matched and one that never existed. So this will
-         * probably never return an error if the user tries to cancel a request that was not
-         * previously started. */
-        MPL_DBG_MSG_FMT(MPIR_DBG_PT2PT, VERBOSE,
-                        (MPL_DBG_FDEST, "Request not found. Assuming already cancelled"));
-        goto fn_exit;
-    } else if (ret < 0) {
-        MPIR_ERR_CHKANDJUMP4(ret < 0, mpi_errno, MPI_ERR_OTHER, "**ofid_cancel",
-                             "**ofid_cancel %s %d %s %s", __SHORT_FILE__, __LINE__, __func__,
-                             fi_strerror(-ret));
-    }
+    /* We can't rely on the return code of fi_cancel, assume always successful.
+     * ref: https://github.com/ofiwg/libfabric/issues/7795
+     */
+    fi_cancel((fid_t) MPIDI_OFI_global.ctx[ctx_idx].rx, &(MPIDI_OFI_REQUEST(rreq, context)));
 
     if (is_blocking) {
+        /* progress until the rreq complets, either with cancel-bit set,
+         * or with message received */
         while (!MPIR_cc_is_complete(&rreq->cc)) {
-            mpi_errno = MPIDI_OFI_progress_uninlined(vni);
+            mpi_errno = MPIDI_OFI_progress_uninlined(vci);
             MPIR_ERR_CHECK(mpi_errno);
         }
+    } else {
+        /* run progress once to prevent accumulating cq errors. */
+        mpi_errno = MPIDI_OFI_progress_uninlined(vci);
+        MPIR_ERR_CHECK(mpi_errno);
     }
 
   fn_exit:
