@@ -42,6 +42,21 @@ cvars:
         }                                       \
     } while (0)
 
+MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_gpu_rma_enabled(const void *ptr)
+{
+    if (ENABLE_GPU) {
+        MPL_pointer_attr_t attr;
+        MPIR_GPU_query_pointer_attr(ptr, &attr);
+        if (MPL_gpu_query_pointer_is_dev(ptr, &attr)) {
+            if (MPIDI_OFI_ENABLE_HMEM && MPIR_GPU_query_pointer_is_strict_dev(ptr, &attr)) {
+                return 1;
+            } else
+                return 0;
+        }
+    }
+    return 1;
+}
+
 MPL_STATIC_INLINE_PREFIX void MPIDI_OFI_query_acc_atomic_support(MPI_Datatype dt, int query_type,
                                                                  MPI_Op op,
                                                                  MPIR_Win * win,
@@ -176,7 +191,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_put(const void *origin_addr,
     MPI_Aint origin_true_lb, target_true_lb;
     struct iovec iov;
     struct fi_rma_iov riov;
-    int nic = 0;
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
     MPIR_FUNC_ENTER;
 
@@ -187,8 +202,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_put(const void *origin_addr,
     MPIDI_Datatype_check_contig_size_extent_lb(target_datatype, target_count, target_contig,
                                                target_bytes, target_extent, target_true_lb);
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
 
     /* zero-byte messages */
     if (unlikely(origin_bytes == 0))
@@ -218,49 +233,60 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_put(const void *origin_addr,
     if (unlikely(!target_mr_found))
         goto am_fallback;
 
+    void *origin_ptr = MPIR_get_contig_ptr(origin_addr, origin_true_lb);
+    MPL_pointer_attr_t attr;
+    MPIR_GPU_query_pointer_attr(origin_ptr, &attr);
+
     /* small contiguous messages */
-    if (origin_contig && target_contig && (origin_bytes <= MPIDI_OFI_global.max_buffered_write) &&
-        !MPIR_CVAR_CH4_OFI_DISABLE_INJECT_WRITE) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+    /* skip fi_inject path for GPU messages because this path can be
+     * very slow */
+    if (origin_contig && target_contig &&
+        (origin_bytes <= MPIDI_OFI_global.max_buffered_write &&
+         !MPL_gpu_query_pointer_is_dev(origin_ptr, &attr))) {
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         MPIDI_OFI_win_cntr_incr(win);
         MPIDI_OFI_CALL_RETRY(fi_inject_write(MPIDI_OFI_WIN(win).ep,
                                              MPIR_get_contig_ptr(origin_addr, origin_true_lb),
                                              target_bytes,
-                                             MPIDI_OFI_av_to_phys(addr, nic, vni, vni_target),
+                                             MPIDI_OFI_av_to_phys(addr, nic_target, vci_target),
                                              target_mr.addr + target_true_lb,
-                                             target_mr.mr_key), vni, rdma_inject_write, FALSE);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+                                             target_mr.mr_key), vci, rdma_inject_write);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto null_op_exit;
     }
 
     /* large contiguous messages */
     if (origin_contig && target_contig) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         if (sigreq) {
             MPIDI_OFI_REQUEST_CREATE(*sigreq, MPIR_REQUEST_KIND__RMA, 0);
             flags = FI_COMPLETION | FI_DELIVERY_COMPLETE;
         } else {
             flags = FI_DELIVERY_COMPLETE;
         }
-        msg.desc = NULL;
-        msg.addr = MPIDI_OFI_av_to_phys(addr, nic, vni, vni_target);
+        iov.iov_base = MPIR_get_contig_ptr(origin_addr, origin_true_lb);
+        iov.iov_len = target_bytes;
+
+        void *desc = NULL;
+        if (MPIR_GPU_query_pointer_is_strict_dev(iov.iov_base, &attr))
+            MPIDI_OFI_gpu_rma_register(iov.iov_base, iov.iov_len, &attr, win, nic_target, &desc);
+
+        msg.desc = desc;
+        msg.addr = MPIDI_OFI_av_to_phys(addr, nic_target, vci_target);
         msg.context = NULL;
         msg.data = 0;
         msg.msg_iov = &iov;
         msg.iov_count = 1;
         msg.rma_iov = &riov;
         msg.rma_iov_count = 1;
-        iov.iov_base = MPIR_get_contig_ptr(origin_addr, origin_true_lb);
-        iov.iov_len = target_bytes;
         riov.addr = target_mr.addr + target_true_lb;
         riov.len = target_bytes;
         riov.key = target_mr.mr_key;
         MPIDI_OFI_INIT_CHUNK_CONTEXT(win, sigreq);
-        MPIDI_OFI_CALL_RETRY(fi_writemsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vni, rdma_write,
-                             FALSE);
+        MPIDI_OFI_CALL_RETRY(fi_writemsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vci, rdma_write);
         /* Complete signal request to inform completion to user. */
         MPIDI_OFI_sigreq_complete(sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
@@ -271,22 +297,22 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_put(const void *origin_addr,
 
     if (origin_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN &&
         target_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         mpi_errno =
             MPIDI_OFI_nopack_putget(origin_addr, origin_count, origin_datatype, target_rank,
                                     target_count, target_datatype, target_mr, win, addr,
                                     MPIDI_OFI_PUT, sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
     if (origin_density < MPIR_CVAR_CH4_IOV_DENSITY_MIN &&
         target_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         mpi_errno =
             MPIDI_OFI_pack_put(origin_addr, origin_count, origin_datatype, target_rank,
                                target_count, target_datatype, target_mr, win, addr, sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
@@ -326,7 +352,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_put(const void *origin_addr,
     int mpi_errno = MPI_SUCCESS;
 
     if (!MPIDI_OFI_ENABLE_RMA || !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-        MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+        !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno = MPIDIG_mpi_put(origin_addr, origin_count, origin_datatype, target_rank,
                                    target_disp, target_count, target_datatype, win);
         goto fn_exit;
@@ -362,7 +389,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get(void *origin_addr,
     MPI_Aint target_bytes, target_extent;
     struct fi_rma_iov riov;
     struct iovec iov;
-    int nic = 0;
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
     MPIR_FUNC_ENTER;
 
@@ -372,8 +399,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get(void *origin_addr,
     MPIDI_Datatype_check_contig_size_extent_lb(target_datatype, target_count, target_contig,
                                                target_bytes, target_extent, target_true_lb);
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
 
     /* zero-byte messages */
     if (unlikely(target_bytes == 0))
@@ -404,32 +431,38 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get(void *origin_addr,
 
     /* contiguous messages */
     if (origin_contig && target_contig) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         if (sigreq) {
             MPIDI_OFI_REQUEST_CREATE(*sigreq, MPIR_REQUEST_KIND__RMA, 0);
             flags = FI_COMPLETION | FI_DELIVERY_COMPLETE;
         } else {
             flags = 0;
         }
-        msg.desc = NULL;
+        iov.iov_base = MPIR_get_contig_ptr(origin_addr, origin_true_lb);
+        iov.iov_len = target_bytes;
+
+        void *desc = NULL;
+        MPL_pointer_attr_t attr;
+        MPIR_GPU_query_pointer_attr(iov.iov_base, &attr);
+        if (MPIR_GPU_query_pointer_is_strict_dev(iov.iov_base, &attr))
+            MPIDI_OFI_gpu_rma_register(iov.iov_base, iov.iov_len, NULL, win, nic_target, &desc);
+
+        msg.desc = desc;
         msg.msg_iov = &iov;
         msg.iov_count = 1;
-        msg.addr = MPIDI_OFI_av_to_phys(addr, nic, vni, vni_target);
+        msg.addr = MPIDI_OFI_av_to_phys(addr, nic_target, vci_target);
         msg.rma_iov = &riov;
         msg.rma_iov_count = 1;
         msg.context = NULL;
         msg.data = 0;
-        iov.iov_base = MPIR_get_contig_ptr(origin_addr, origin_true_lb);
-        iov.iov_len = target_bytes;
         riov.addr = target_mr.addr + target_true_lb;
         riov.len = target_bytes;
         riov.key = target_mr.mr_key;
         MPIDI_OFI_INIT_CHUNK_CONTEXT(win, sigreq);
-        MPIDI_OFI_CALL_RETRY(fi_readmsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vni, rdma_write,
-                             FALSE);
+        MPIDI_OFI_CALL_RETRY(fi_readmsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vci, rdma_write);
         /* Complete signal request to inform completion to user. */
         MPIDI_OFI_sigreq_complete(sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
@@ -440,22 +473,22 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get(void *origin_addr,
 
     if (origin_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN &&
         target_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         mpi_errno =
             MPIDI_OFI_nopack_putget(origin_addr, origin_count, origin_datatype, target_rank,
                                     target_count, target_datatype, target_mr, win, addr,
                                     MPIDI_OFI_GET, sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
     if (origin_density < MPIR_CVAR_CH4_IOV_DENSITY_MIN &&
         target_density >= MPIR_CVAR_CH4_IOV_DENSITY_MIN) {
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         mpi_errno =
             MPIDI_OFI_pack_get(origin_addr, origin_count, origin_datatype, target_rank,
                                target_count, target_datatype, target_mr, win, addr, sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
@@ -495,7 +528,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_get(void *origin_addr,
     MPIR_FUNC_ENTER;
 
     if (!MPIDI_OFI_ENABLE_RMA || !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-        MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+        !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno = MPIDIG_mpi_get(origin_addr, origin_count, origin_datatype, target_rank,
                                    target_disp, target_count, target_datatype, win);
         goto fn_exit;
@@ -527,7 +561,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_rput(const void *origin_addr,
     int mpi_errno = MPI_SUCCESS;
 
     if (!MPIDI_OFI_ENABLE_RMA || !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-        MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+        !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno = MPIDIG_mpi_rput(origin_addr, origin_count, origin_datatype, target_rank,
                                     target_disp, target_count, target_datatype, win, request);
         goto fn_exit;
@@ -565,10 +600,10 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_compare_and_swap(const void *origin_ad
     struct fi_ioc comparev;
     struct fi_rma_ioc targetv;
     struct fi_msg_atomic msg;
-    int nic = 0;
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
 
     if (
 #ifndef MPIDI_CH4_DIRECT_NETMOD
@@ -581,9 +616,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_compare_and_swap(const void *origin_ad
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
            !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-           MPIR_GPU_query_pointer_is_dev(origin_addr) ||
-           MPIR_GPU_query_pointer_is_dev(compare_addr) ||
-           MPIR_GPU_query_pointer_is_dev(result_addr)) {
+           !MPIDI_OFI_gpu_rma_enabled(origin_addr) ||
+           !MPIDI_OFI_gpu_rma_enabled(compare_addr) || !MPIDI_OFI_gpu_rma_enabled(result_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_compare_and_swap(origin_addr, compare_addr, result_addr, datatype,
                                         target_rank, target_disp, win);
@@ -633,10 +668,17 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_compare_and_swap(const void *origin_ad
     targetv.count = 1;
     targetv.key = target_mr.mr_key;
 
+    void *desc = NULL;
+    MPIDI_OFI_gpu_rma_register(originv.addr, dt_size, NULL, win, nic_target, &desc);
+    void *compare_desc = NULL;
+    MPIDI_OFI_gpu_rma_register(comparev.addr, dt_size, NULL, win, nic_target, &compare_desc);
+    void *result_desc = NULL;
+    MPIDI_OFI_gpu_rma_register(resultv.addr, dt_size, NULL, win, nic_target, &result_desc);
+
     msg.msg_iov = &originv;
-    msg.desc = NULL;
+    msg.desc = desc;
     msg.iov_count = 1;
-    msg.addr = MPIDI_OFI_av_to_phys(av, nic, vni, vni_target);
+    msg.addr = MPIDI_OFI_av_to_phys(av, nic_target, vci_target);
     msg.rma_iov = &targetv;
     msg.rma_iov_count = 1;
     msg.datatype = fi_dt;
@@ -644,12 +686,12 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_compare_and_swap(const void *origin_ad
     msg.context = NULL;
     msg.data = 0;
 
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
     MPIDI_OFI_win_cntr_incr(win);
     MPIDI_OFI_CALL_RETRY(fi_compare_atomicmsg(MPIDI_OFI_WIN(win).ep, &msg,
-                                              &comparev, NULL, 1, &resultv, NULL, 1, 0), vni,
-                         atomicto, FALSE);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+                                              &comparev, compare_desc, 1, &resultv, result_desc, 1,
+                                              0), vci, atomicto);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
   fn_exit:
     MPIR_FUNC_EXIT;
     return mpi_errno;
@@ -658,9 +700,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_compare_and_swap(const void *origin_ad
   am_fallback:
     /* Wait for OFI case to complete for atomicity.
      * For now, there is no FI flag to track atomic only ops, we use RMA level cntr. */
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
-    MPIDI_OFI_win_do_progress(win, vni);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
+    MPIDI_OFI_win_do_progress(win, vci);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
     return MPIDIG_mpi_compare_and_swap(origin_addr, compare_addr, result_addr, datatype,
                                        target_rank, target_disp, win);
 }
@@ -681,7 +723,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_accumulate(const void *origin_addr,
     int target_contig, origin_contig;
     MPI_Aint target_bytes, origin_bytes, target_extent;
     MPI_Aint origin_true_lb, target_true_lb;
-    int nic = 0;
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
     MPIR_FUNC_ENTER;
 
@@ -694,8 +736,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_accumulate(const void *origin_addr,
     if (origin_bytes == 0)
         goto null_op_exit;
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
 
     /* prepare remote addr and mr key.
      * Continue native path only when all segments are in the same registered memory region */
@@ -736,7 +778,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_accumulate(const void *origin_addr,
         /* Ensure completion of outstanding AMs for atomicity. */
         MPIDIG_wait_am_acc(win, target_rank);
 
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         uint64_t flags;
         if (sigreq) {
             MPIDI_OFI_REQUEST_CREATE(*sigreq, MPIR_REQUEST_KIND__RMA, 0);
@@ -755,10 +797,14 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_accumulate(const void *origin_addr,
         targetv.count = basic_count;
         targetv.key = target_mr.mr_key;
 
+        void *desc = NULL;
+        MPIDI_OFI_gpu_rma_register(originv.addr, dt_size * basic_count, NULL, win, nic_target,
+                                   &desc);
+
         msg.msg_iov = &originv;
-        msg.desc = NULL;
+        msg.desc = desc;
         msg.iov_count = 1;
-        msg.addr = MPIDI_OFI_av_to_phys(addr, nic, vni, vni_target);
+        msg.addr = MPIDI_OFI_av_to_phys(addr, nic_target, vci_target);
         msg.rma_iov = &targetv;
         msg.rma_iov_count = 1;
         msg.datatype = fi_dt;
@@ -766,20 +812,19 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_accumulate(const void *origin_addr,
         msg.context = NULL;
         msg.data = 0;
         MPIDI_OFI_INIT_CHUNK_CONTEXT(win, sigreq);
-        MPIDI_OFI_CALL_RETRY(fi_atomicmsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vni,
-                             rdma_atomicto, FALSE);
+        MPIDI_OFI_CALL_RETRY(fi_atomicmsg(MPIDI_OFI_WIN(win).ep, &msg, flags), vci, rdma_atomicto);
         /* Complete signal request to inform completion to user. */
         MPIDI_OFI_sigreq_complete(sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
   am_fallback:
     /* Wait for OFI acc to complete for atomicity.
      * For now, there is no FI flag to track atomic only ops, we use RMA level cntr. */
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
-    MPIDI_OFI_win_do_progress(win, vni);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
+    MPIDI_OFI_win_do_progress(win, vci);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
     if (sigreq)
         mpi_errno = MPIDIG_mpi_raccumulate(origin_addr, origin_count, origin_datatype, target_rank,
                                            target_disp, target_count, target_datatype, op, win,
@@ -819,7 +864,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get_accumulate(const void *origin_addr
     int target_contig, origin_contig, result_contig;
     MPI_Aint target_bytes, target_extent;
     MPI_Aint origin_true_lb, target_true_lb, result_true_lb;
-    int nic = 0;
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
     MPIR_FUNC_ENTER;
 
@@ -832,8 +877,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get_accumulate(const void *origin_addr
     if (target_bytes == 0)
         goto null_op_exit;
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
 
     /* contiguous messages */
     if (origin_contig && target_contig && result_contig) {
@@ -874,7 +919,7 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get_accumulate(const void *origin_addr
         /* Ensure completion of outstanding AMs for atomicity. */
         MPIDIG_wait_am_acc(win, target_rank);
 
-        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
         uint64_t flags;
         if (sigreq) {
             MPIDI_OFI_REQUEST_CREATE(*sigreq, MPIR_REQUEST_KIND__RMA, 0);
@@ -895,10 +940,17 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get_accumulate(const void *origin_addr
         targetv.count = basic_count;
         targetv.key = target_mr.mr_key;
 
+        void *desc = NULL;
+        MPIDI_OFI_gpu_rma_register(originv.addr, dt_size * basic_count, NULL, win, nic_target,
+                                   &desc);
+        void *result_desc = NULL;
+        MPIDI_OFI_gpu_rma_register(resultv.addr, dt_size * basic_count, NULL, win, nic_target,
+                                   &result_desc);
+
         msg.msg_iov = &originv;
-        msg.desc = NULL;
+        msg.desc = desc;
         msg.iov_count = 1;
-        msg.addr = MPIDI_OFI_av_to_phys(addr, nic, vni, vni_target);
+        msg.addr = MPIDI_OFI_av_to_phys(addr, nic_target, vci_target);
         msg.rma_iov = &targetv;
         msg.rma_iov_count = 1;
         msg.datatype = fi_dt;
@@ -907,19 +959,19 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_OFI_do_get_accumulate(const void *origin_addr
         msg.data = 0;
         MPIDI_OFI_INIT_CHUNK_CONTEXT(win, sigreq);
         MPIDI_OFI_CALL_RETRY(fi_fetch_atomicmsg(MPIDI_OFI_WIN(win).ep, &msg, &resultv,
-                                                NULL, 1, flags), vni, rdma_readfrom, FALSE);
+                                                result_desc, 1, flags), vci, rdma_readfrom);
         /* Complete signal request to inform completion to user. */
         MPIDI_OFI_sigreq_complete(sigreq);
-        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+        MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
         goto fn_exit;
     }
 
   am_fallback:
     /* Wait for OFI getacc to complete for atomicity.
      * For now, there is no FI flag to track atomic only ops, we use RMA level cntr. */
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
-    MPIDI_OFI_win_do_progress(win, vni);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
+    MPIDI_OFI_win_do_progress(win, vci);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
     if (sigreq)
         mpi_errno =
             MPIDIG_mpi_rget_accumulate(origin_addr, origin_count, origin_datatype, result_addr,
@@ -970,7 +1022,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_raccumulate(const void *origin_addr,
            !(winattr & MPIDI_WINATTR_ACCU_NO_SHM) ||
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
-           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_raccumulate(origin_addr, origin_count, origin_datatype, target_rank,
                                    target_disp, target_count, target_datatype, op, win, request);
@@ -1019,8 +1072,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_rget_accumulate(const void *origin_add
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
            !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-           MPIR_GPU_query_pointer_is_dev(origin_addr) ||
-           MPIR_GPU_query_pointer_is_dev(result_addr)) {
+           !MPIDI_OFI_gpu_rma_enabled(origin_addr) || !MPIDI_OFI_gpu_rma_enabled(result_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_rget_accumulate(origin_addr, origin_count, origin_datatype, result_addr,
                                        result_count, result_datatype, target_rank, target_disp,
@@ -1058,9 +1111,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_fetch_and_op(const void *origin_addr,
     struct fi_msg_atomic msg;
     MPIR_FUNC_ENTER;
 
-    int vni = MPIDI_WIN(win, am_vci);
-    int vni_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
-    int nic = 0;
+    int vci = MPIDI_WIN(win, am_vci);
+    int vci_target = MPIDI_WIN_TARGET_VCI(win, target_rank);
+    int nic_target = MPIDI_OFI_get_pref_nic(win->comm_ptr, target_rank);
 
     if (
 #ifndef MPIDI_CH4_DIRECT_NETMOD
@@ -1073,8 +1126,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_fetch_and_op(const void *origin_addr,
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
            !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-           MPIR_GPU_query_pointer_is_dev(origin_addr) ||
-           MPIR_GPU_query_pointer_is_dev(result_addr)) {
+           !MPIDI_OFI_gpu_rma_enabled(origin_addr) || !MPIDI_OFI_gpu_rma_enabled(result_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_fetch_and_op(origin_addr, result_addr, datatype, target_rank, target_disp,
                                     op, win);
@@ -1120,10 +1173,15 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_fetch_and_op(const void *origin_addr,
     targetv.count = 1;
     targetv.key = target_mr.mr_key;
 
+    void *desc = NULL;
+    MPIDI_OFI_gpu_rma_register(originv.addr, dt_size, NULL, win, nic_target, &desc);
+    void *result_desc = NULL;
+    MPIDI_OFI_gpu_rma_register(resultv.addr, dt_size, NULL, win, nic_target, &result_desc);
+
     msg.msg_iov = &originv;
-    msg.desc = NULL;
+    msg.desc = desc;
     msg.iov_count = 1;
-    msg.addr = MPIDI_OFI_av_to_phys(av, nic, vni, vni_target);
+    msg.addr = MPIDI_OFI_av_to_phys(av, nic_target, vci_target);
     msg.rma_iov = &targetv;
     msg.rma_iov_count = 1;
     msg.datatype = fi_dt;
@@ -1131,11 +1189,11 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_fetch_and_op(const void *origin_addr,
     msg.context = NULL;
     msg.data = 0;
 
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
     MPIDI_OFI_win_cntr_incr(win);
     MPIDI_OFI_CALL_RETRY(fi_fetch_atomicmsg(MPIDI_OFI_WIN(win).ep, &msg, &resultv,
-                                            NULL, 1, 0), vni, rdma_readfrom, FALSE);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+                                            result_desc, 1, 0), vci, rdma_readfrom);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
 
   fn_exit:
     MPIR_FUNC_EXIT;
@@ -1145,9 +1203,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_fetch_and_op(const void *origin_addr,
   am_fallback:
     /* Wait for OFI fetch_and_op to complete for atomicity.
      * For now, there is no FI flag to track atomic only ops, we use RMA level cntr. */
-    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vni).lock);
-    MPIDI_OFI_win_do_progress(win, vni);
-    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vni).lock);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(vci).lock);
+    MPIDI_OFI_win_do_progress(win, vci);
+    MPID_THREAD_CS_EXIT(VCI, MPIDI_VCI(vci).lock);
     return MPIDIG_mpi_fetch_and_op(origin_addr, result_addr, datatype, target_rank, target_disp, op,
                                    win);
 }
@@ -1166,7 +1224,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_rget(void *origin_addr,
     int mpi_errno = MPI_SUCCESS;
 
     if (!MPIDI_OFI_ENABLE_RMA || !(winattr & MPIDI_WINATTR_NM_REACHABLE) ||
-        MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+        !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno = MPIDIG_mpi_rget(origin_addr, origin_count, origin_datatype, target_rank,
                                     target_disp, target_count, target_datatype, win, request);
         goto fn_exit;
@@ -1211,8 +1270,9 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_get_accumulate(const void *origin_addr
            !(winattr & MPIDI_WINATTR_ACCU_NO_SHM) ||
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
-           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || MPIR_GPU_query_pointer_is_dev(origin_addr) ||
-           MPIR_GPU_query_pointer_is_dev(result_addr)) {
+           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || !MPIDI_OFI_gpu_rma_enabled(origin_addr) ||
+           !MPIDI_OFI_gpu_rma_enabled(result_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_get_accumulate(origin_addr, origin_count, origin_datatype, result_addr,
                                       result_count, result_datatype, target_rank, target_disp,
@@ -1253,7 +1313,8 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_NM_mpi_accumulate(const void *origin_addr,
            !(winattr & MPIDI_WINATTR_ACCU_NO_SHM) ||
 #endif
            !MPIDI_OFI_ENABLE_RMA || !MPIDI_OFI_ENABLE_ATOMICS ||
-           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || MPIR_GPU_query_pointer_is_dev(origin_addr)) {
+           !(winattr & MPIDI_WINATTR_NM_REACHABLE) || !MPIDI_OFI_gpu_rma_enabled(origin_addr)) {
+        MPIDI_OFI_register_am_bufs();
         mpi_errno =
             MPIDIG_mpi_accumulate(origin_addr, origin_count, origin_datatype, target_rank,
                                   target_disp, target_count, target_datatype, op, win);
