@@ -1,19 +1,20 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *  (C) 2014 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <assert.h>
 #include "mpi.h"
+#include "mpitest.h"
 
 int main(int argc, char **argv)
 {
     MPI_Request req;
     MPI_Status status;
 
-    MPI_Init(NULL, NULL);
+    MTest_Init(&argc, &argv);
 
     MPI_Issend(NULL, 0, MPI_BYTE, 0, 123, MPI_COMM_SELF, &req);
 
@@ -30,7 +31,7 @@ int main(int argc, char **argv)
     MPI_Cancel(&req);
     MPI_Wait(&req, &status);
 
-    printf(" No Errors\n");
+    MTest_Finalize(0);
 
-    MPI_Finalize();
+    return 0;
 }
